//
// Fan Control
// Copyright 2006 Lobotomo Software 
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//

#import "MFDaemon.h"
#import "MFProtocol.h"
#import "MFDefinitions.h"
#import "smc.h"

#define MFApplicationIdentifier     "com.lobotomo.MacProFan"


@implementation MFDaemon

- (id)init
{
    if (self = [super init]) {
        // set sane defaults
        lowerThreshold = 25.0;
        upperThreshold = 50.0;
        baseRpm = 1000;
        maxRpm = MFMaxRpm;
    }
    return self;
}

// store preferences
- (void)storePreferences
{
    CFPreferencesSetValue(CFSTR("baseRpm"), (CFPropertyListRef)[NSNumber numberWithInt:baseRpm],
                          CFSTR(MFApplicationIdentifier), kCFPreferencesAnyUser, kCFPreferencesCurrentHost);
    CFPreferencesSetValue(CFSTR("lowerThreshold"), (CFPropertyListRef)[NSNumber numberWithFloat:lowerThreshold],
                          CFSTR(MFApplicationIdentifier), kCFPreferencesAnyUser, kCFPreferencesCurrentHost);
    CFPreferencesSetValue(CFSTR("upperThreshold"), (CFPropertyListRef)[NSNumber numberWithFloat:upperThreshold],
                          CFSTR(MFApplicationIdentifier), kCFPreferencesAnyUser, kCFPreferencesCurrentHost);
    CFPreferencesSetValue(CFSTR("fahrenheit"), (CFPropertyListRef)[NSNumber numberWithBool:fahrenheit],
                          CFSTR(MFApplicationIdentifier), kCFPreferencesAnyUser, kCFPreferencesCurrentHost);
    CFPreferencesSynchronize(CFSTR(MFApplicationIdentifier), kCFPreferencesAnyUser, kCFPreferencesCurrentHost);
}

// read preferences
- (void)readPreferences
{
    CFPropertyListRef property;
    property = CFPreferencesCopyValue(CFSTR("baseRpm"), CFSTR(MFApplicationIdentifier),
               kCFPreferencesAnyUser, kCFPreferencesCurrentHost);
    if (property) {
        baseRpm = [(NSNumber *)property intValue];
    }
    property = CFPreferencesCopyValue(CFSTR("lowerThreshold"), CFSTR(MFApplicationIdentifier),
               kCFPreferencesAnyUser, kCFPreferencesCurrentHost);
    if (property) {
        lowerThreshold = [(NSNumber *)property floatValue];
    }
    property = CFPreferencesCopyValue(CFSTR("upperThreshold"), CFSTR(MFApplicationIdentifier),
               kCFPreferencesAnyUser, kCFPreferencesCurrentHost);
    if (property) {
        upperThreshold = [(NSNumber *)property floatValue];
    }
    property = CFPreferencesCopyValue(CFSTR("fahrenheit"), CFSTR(MFApplicationIdentifier),
               kCFPreferencesAnyUser, kCFPreferencesCurrentHost);
    if (property) {
        fahrenheit = [(NSNumber *)property boolValue];
    }
}

// this gets called after application start
- (void)start
{
	// unlike the other fans the PS Fan needs to be put into forced more in order to respond to rpm change request.
	SMCVal_t	val;

	SMCOpen();
    strcpy(val.key, SMC_KEY_FAN_CONTROL_MODE);
    val.bytes[0] = 0;
    val.bytes[1] = SMC_KEY_FAN3_MASK;
    val.dataSize = 2;
	SMCWriteKey(val);									// write out the new fan control key value
	SMCClose();
	
    [self readPreferences];
    [NSTimer scheduledTimerWithTimeInterval:5.0 target:self selector:@selector(timer:) userInfo:nil repeats:YES];
}

// control loop called by NSTimer
- (void)timer:(NSTimer *)aTimer
{
    double temp;
	double t0, t1;
    int targetRpm;
    int step;

    SMCOpen();

	// read the 2 heatsink temps
    t0 = SMCGetTemperature(SMC_KEY_HEATSINK0_TEMP);
    t1 = SMCGetTemperature(SMC_KEY_HEATSINK1_TEMP);

	// get the max temp and use it
	temp = t0;
	if( t1> temp ) temp = t1;

 
    if (temp < lowerThreshold) {
        targetRpm = baseRpm;
    } else if (temp > upperThreshold) {
        targetRpm = maxRpm;
    } else {
        targetRpm = baseRpm + (temp - lowerThreshold) / (upperThreshold - lowerThreshold) * (maxRpm - baseRpm);
    }

    // adjust fan speed in reasonable steps - no need to be too dynamic
    if (currentRpm == 0) {
        step = targetRpm;
    } else {
        step = (targetRpm - currentRpm) / 6;
        if (abs(step) < 20) {
            step = targetRpm - currentRpm;
        }
    }
    targetRpm = currentRpm = currentRpm + step;
    
    SMCSetFanRpm(SMC_KEY_FAN0_RPM_MIN, targetRpm);	// CPU/Memory Intake Front fan
    SMCSetFanRpm(SMC_KEY_FAN1_RPM_MIN, targetRpm);	// PCIe expansion area front fan
    SMCSetFanRpm(SMC_KEY_FAN2_RPM_MIN, targetRpm + 200);	// CPU/Memory Exhaust back fan
    SMCSetFanRpm(SMC_KEY_FAN3_RPM_TGT, targetRpm + 200);	// Power Supply Exhaust back fan

    SMCClose();

    // save preferences
    if (needWrite) {
        [self storePreferences];
        needWrite = NO;
    }
}

// accessors
- (int)baseRpm
{
	return baseRpm;
}

- (void)setBaseRpm:(int)newBaseRpm
{
	baseRpm = newBaseRpm;
    needWrite = YES;
}

- (float)lowerThreshold
{
	return lowerThreshold;
}

- (void)setLowerThreshold:(float)newLowerThreshold
{
	lowerThreshold = newLowerThreshold;
    needWrite = YES;
}

- (float)upperThreshold
{
	return upperThreshold;
}

- (void)setUpperThreshold:(float)newUpperThreshold
{
	upperThreshold = newUpperThreshold;
    needWrite = YES;
}

- (BOOL)fahrenheit
{
	return fahrenheit;
}

- (void)setFahrenheit:(BOOL)newFahrenheit
{
	fahrenheit = newFahrenheit;
    needWrite = YES;
}

- (void)temperature:(float *)temperature leftFanRpm:(int *)leftFanRpm rightFanRpm:(int *)rightFanRpm
{
	double temp, t0, t1;

    SMCOpen();
    if (temperature) {
		// read the 2 heat sink temps
        t0 = SMCGetTemperature(SMC_KEY_HEATSINK0_TEMP);
        t1 = SMCGetTemperature(SMC_KEY_HEATSINK1_TEMP);

		// get the max temp and use it
		temp = t0;
		if( t1> temp ) temp = t1;

         *temperature = temp;
    }
    if (leftFanRpm) {
        *leftFanRpm = SMCGetFanRpm(SMC_KEY_FAN0_RPM_MIN);
    }
    if (rightFanRpm) {
        *rightFanRpm = SMCGetFanRpm(SMC_KEY_FAN1_RPM_MIN);
    }
    SMCClose();
}

@end
